/* AVR  High voltage Chip Eraser
 *
 *
 *                         ATtiny2313
 *
 *                      ~RESET      VCC
 *       TGT_VCC_CTRL      PD0      PB7  D7
 *       TGT_RESET_CTRL    PD1      PB6  D6
 *       TGT_XTAL1         PA1      PB5  D5
 *       TGT_OE            PA0      PB4  D4
 *       TGT_WR            PD2      PB3  D3 / HSV_MISO  (SII)
 *       TGT_BS1           PD3      PB2  D2 / HSV_SCK   (SDO)
 *       TGT_XA0           PD4      PB1  D1 / HSV_MOSI  (SDI)
 *       PWM               PD5      PB0  D0 / HSV_XTAL1 (SCI)
 *                         GND      PD6  TGT_XA1 (/BS2)
 *
 *
 * ATtiny2313 Fuses  L: 0x64 H: 0xdf -> 0xdd (Brown out @2.7V)
 *           
 *   VCC ------------- PNP ---+---> TGT_VCC          
 *                      |     |
 *  TGT_VCC_CTRL -------+     C(0.1u)
 *                            |
 *                         ---+---
 *
 *  +12V ------------- PNP ---+--> TGT_RESET          
 *                      |     |
 *  TGT_RESET_CTL  --- NPN    R (10K)
 *                      |     |
 *                    --+-----+---
 *
 *
 * TGT_XTAL1 --- GREEN_LED --- 150 ---- GND
 * TGT_OE    --- RED_LED ----- 220 ---- GND
 *
 * TAGETS:
 *
 * ATtiny45 Type
 *       
 *            TGT_RESET      TGT_VCC
 *            HSV_XTAL1      HSV_SCK
 *                   NC      HSV_MISO
 *                  GND      HSV_MOSI
 *       
 * ATtiny44 Type
 *       
 *              TGT_VCC      GND
 *            HSV_XTAL1      NC
 *                   NC      NC
 *            TGT_RESET      NC
 *                   NC      NC
 *                   NC      HSV_SCK
 *             HSV_MOSI      HSV_MISO
 *       
 * ATtiny2313 Type
 *       
 *            TGT_RESET      TGT_VCC
 *                   NC      D7
 *                   NC      D6
 *                   NC      D5
 *            TGT_XTAL1      D4
 *               TGT_OE      D3
 *               TGT_WR      D2
 *              TGT_BS1      D1
 *              TGT_XA0      D0
 *                  GND      TGT_XA1
 *       
 *  ATtiny461 Type
 *       
 *               TGT_WR      D0
 *              TGT_XA0      D1
 *              TGT_XA1      D2
 *              TGT_BS1      D3
 *              TGT_VCC      GND
 *                  GND      NC
 *            TGT_XTAL1      D4
 *               TGT_OE      D5
 *                   NC      D6
 *            TGT_RESET      D7
 *       
 * ATmega88 Type
 *       
 *            TGT_RESET      NC
 *                   NC      NC
 *                   NC      NC
 *               TGT_OE      TGT_BS2
 *               TGT_WR      D7
 *              TGT_BS1      D6
 *              TGT_VCC      GND
 *                  GND      NC
 *            TGT_XTAL1      NC
 *                   NC      D5
 *              TGT_XA0      D4
 *    TGT_BS2 ----  XA1      D3
 *    GND -- R -- PAGEL      D2
 *                   D0      D1
 *
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <util/delay.h>

//#define TGT_VCC_DIRECT
//#define NO_HSV
//#define NO_PP

#define bit_set(port,bit)       { port |= (1<<(bit));}
#define bit_clear(port,bit)     { port &= ~(1<<(bit));}
#define nop()   __asm__ volatile ("nop" )

static void ss_sleep(uint8_t sub_sec) {
	while (sub_sec--) {
	    uint16_t loop_cnt = 3333;
	    while (loop_cnt--) {
		_delay_loop_1((uint8_t)(F_CPU/100000)); // 30us, 80 @ 8Mhz
	    }
	}
}

static void mdelay(uint8_t msec) {
	while (msec--) {
	    uint8_t loop_cnt = 33;
	    while (loop_cnt--) {
		_delay_loop_1((uint8_t)(F_CPU/100000)); // 30us, 80 @ 8Mhz
	    }
	}
}

#define PWM_DDR		DDRD
#define PWM_PORT	PORTD
#define PWM		PD5

#define TGT_XTAL1_PORT	PORTA
#define TGT_XTAL1_DDR	DDRA

#define TGT_OE_PORT	PORTA
#define TGT_OE_DDR	DDRA

#define TGT_XTAL1	PA1
#define TGT_OE		PA0

#define TGT_PORT	PORTD
#define TGT_DDR		DDRD
#define TGT_PIN		PIND

#define TGT_VCC_CTRL	PD0
#define TGT_RESET_CTRL	PD1

#define TGT_WR		PD2
#define TGT_BS1		PD3
#define TGT_XA0		PD4
#define TGT_XA1		PD6
#define TGT_BS2		TGT_XA1

#define TGT_BUS_PORT	PORTB
#define TGT_BUS_DDR	DDRB
#define TGT_BUS_PIN	PINB

#define HSV_PORT	PORTB
#define HSV_DDR		DDRB
#define HSV_PIN		PINB

#define HSV_MISO	PB3
#define HSV_SCK		PB2
#define HSV_MOSI	PB1
#define HSV_XTAL1	PB0

#define HSV_SII	HSV_MISO
#define HSV_SDO	HSV_SCK
#define HSV_SDI	HSV_MOSI
#define HSV_SCI	HSV_XTAL1

static inline void set_vcc(uint8_t on) {
#ifdef TGT_VCC_DIRECT
	on = !on;
#endif
	if (on) {
		bit_clear(TGT_PORT,TGT_VCC_CTRL);
	} else {
		bit_set(TGT_PORT,TGT_VCC_CTRL);
	}
}

static inline void set_reset(uint8_t on) {
	if (on) {
		bit_set(TGT_PORT,TGT_RESET_CTRL);
	} else {
		bit_clear(TGT_PORT,TGT_RESET_CTRL);
	}
}

static inline void init_port() {
	set_vcc(0);
	TGT_DDR = (1<<TGT_VCC_CTRL) | (1<<TGT_RESET_CTRL);
}

static inline void pwm_start() {
	bit_set(PWM_DDR, PWM);
	OCR0A = 32 - 1; // top : 250 kHz (8Mhz/32)
	OCR0B = 16 - 1; // duty 50 %
	TCCR0A = (2<<COM0B0) | (3<<WGM00);
	TCCR0B = (1<<WGM02) | 1; // timer0 start
}

static inline void pwm_stop() {
	bit_clear(PWM_PORT, PWM);
	bit_clear(PWM_DDR, PWM);
	TCCR0B = 0;
}

static inline void program_disable() {
	set_vcc(0);
	set_reset(0);
	pwm_stop();
	bit_clear(TGT_XTAL1_PORT, TGT_XTAL1);
	bit_clear(TGT_OE_PORT, TGT_OE);
	mdelay(100);
	set_vcc(1);
}

static inline void program_enable() {
	uint8_t i;

	bit_set(TGT_XTAL1_DDR, TGT_XTAL1);
	bit_set(TGT_OE_DDR, TGT_OE);
	TGT_DDR |= (1<<TGT_WR) | (1<<TGT_BS1) | (1<<TGT_XA0) | (1<<TGT_BS2);
	HSV_DDR = (1<<HSV_SDI) | (1<<HSV_SII) | (1<<HSV_SDO) | (1<<HSV_SCI);

	pwm_start();
	mdelay(100);

	/* VCC ON */
	bit_set(TGT_OE_PORT, TGT_OE);
	set_vcc(1);
	for (i=0; i<6; i++) {
		nop();
		bit_set(TGT_XTAL1_PORT, TGT_XTAL1);
		nop();
		bit_clear(TGT_XTAL1_PORT, TGT_XTAL1);
	}
	/* TGT_RESET 12V */
	set_reset(1);
	_delay_loop_1((uint8_t)(F_CPU/50000)); // 60 us
	bit_clear(HSV_DDR, HSV_SDO);
	bit_set(TGT_PORT, TGT_WR);
	mdelay(1);
}

#define OP_IDLE		0
#define OP_CHIP_ERASE	1
#define OP_SIGNATURE	2
#define OP_RD_FUSE	3
#define OP_RD_FUSE_L	(OP_RD_FUSE+0)
#define OP_RD_FUSE_H	(OP_RD_FUSE+1)
#define OP_RD_FUSE_Ext	(OP_RD_FUSE+2)
#define OP_WR_FUSE	6
#define OP_WR_FUSE_L	(OP_WR_FUSE+0)
#define OP_WR_FUSE_H	(OP_WR_FUSE+1)
#define OP_WR_FUSE_Ext	(OP_WR_FUSE+2)

#ifndef NO_HSV
static void hsv_clock() {
	_delay_loop_1((uint8_t)(F_CPU/1000000) *  5); // 15 us
	bit_set(HSV_PORT, HSV_SCI);
	_delay_loop_1((uint8_t)(F_CPU/1000000) * 10); // 30 us
	bit_clear(HSV_PORT, HSV_SCI);
	_delay_loop_1((uint8_t)(F_CPU/1000000) *  5); // 15 us
}

static uint8_t hsv_write(uint8_t sii, uint8_t sdi) {
	uint8_t sdo = 0;
	uint8_t i;
	/* send 0,0 first */
	bit_clear(HSV_PORT, HSV_SII);
	bit_clear(HSV_PORT, HSV_SDI);
	hsv_clock();
	for (i=0; i<8; i++) {
		sdo <<= 1;
		if (bit_is_set(HSV_PIN, HSV_SDO)) {
			sdo |= 0x1;
		}
		if (sii & 0x80) {
			bit_set(HSV_PORT, HSV_SII);
		} else {
			bit_clear(HSV_PORT, HSV_SII);
		}
		if (sdi & 0x80) {
			bit_set(HSV_PORT, HSV_SDI);
		} else {
			bit_clear(HSV_PORT, HSV_SDI);
		}
		sdi <<= 1;
		sii <<= 1;
		hsv_clock();
	}
	/* send 00,00 last */
	bit_clear(HSV_PORT, HSV_SII);
	bit_clear(HSV_PORT, HSV_SDI);
	hsv_clock();
	hsv_clock();
	return sdo;
}

static uint8_t hsv_op(uint8_t op, uint8_t data) {
	uint8_t r;
	r = 0xff;
	mdelay(1);
	if (op == OP_IDLE) {
		while (data--) {
			hsv_write(0x4c, 0x00);
		}
	} else if (op == OP_SIGNATURE) {
		hsv_write(0x4c, 0x08);
		hsv_write(0x0c, data);
		hsv_write(0x68, 0x00);
		r = hsv_write(0x6c, 0x00);
	} else if (op == OP_CHIP_ERASE) {
		hsv_write(0x4c, 0x80);
		hsv_write(0x64, 0x00);
		hsv_write(0x6c, 0x00);
		mdelay(100);
		hsv_write(0x4c, 0x00); // null cmd
	} else if (op == OP_RD_FUSE_L) {
		hsv_write(0x4c, 0x04);
		hsv_write(0x68, 0x00);
		r = hsv_write(0x6c, 0x00);
	} else if (op == OP_RD_FUSE_H) {
		hsv_write(0x4c, 0x04);
		hsv_write(0x7a, 0x00);
		r = hsv_write(0x7c, 0x00);
	} else if (op == OP_RD_FUSE_Ext) {
		hsv_write(0x4c, 0x04);
		hsv_write(0x6a, 0x00);
		r = hsv_write(0x6e, 0x00);
	} else if (op == OP_WR_FUSE_L) {
		hsv_write(0x4c, 0x40);
		hsv_write(0x2c, data);
		hsv_write(0x64, 0x00);
		hsv_write(0x6c, 0x00);
		mdelay(100);
		r = 0;
	} else if (op == OP_WR_FUSE_H) {
		hsv_write(0x4c, 0x40);
		hsv_write(0x2c, data);
		hsv_write(0x74, 0x00);
		hsv_write(0x7c, 0x00);
		mdelay(100);
		r = 0;
	} else if (op == OP_WR_FUSE_Ext) {
		hsv_write(0x4c, 0x40);
		hsv_write(0x2c, data);
		hsv_write(0x66, 0x00);
		hsv_write(0x6e, 0x00);
		mdelay(100);
		r = 0;
	}
	return r;
}
#endif


#ifndef NO_PP
static void pp_bus_write(uint8_t data) {
	TGT_BUS_DDR = 0xff;
	TGT_BUS_PORT = data;
	_delay_loop_1((uint8_t)(F_CPU/1000000) *  5); // 15 us
	bit_set(TGT_XTAL1_PORT, TGT_XTAL1);
	_delay_loop_1((uint8_t)(F_CPU/1000000) * 10); // 30 us
	bit_clear(TGT_XTAL1_PORT, TGT_XTAL1);
	_delay_loop_1((uint8_t)(F_CPU/1000000) *  5); // 15 us
	TGT_BUS_DDR = 0;
	TGT_BUS_PORT = 0;
}

static uint8_t pp_bus_read() {
	uint8_t data;
	TGT_BUS_DDR = 0;
	TGT_BUS_PORT = 0;
	_delay_loop_1((uint8_t)(F_CPU/1000000) *  5); // 15 us
	bit_clear(TGT_OE_PORT, TGT_OE);
	_delay_loop_1((uint8_t)(F_CPU/1000000) * 10); // 30 us
	data = TGT_BUS_PIN;
	bit_set(TGT_OE_PORT, TGT_OE);
	_delay_loop_1((uint8_t)(F_CPU/1000000) *  5); // 15 us
	return data;
}

static void pp_cmd(uint8_t cmd) {
	bit_set(TGT_PORT,TGT_XA1);
	pp_bus_write(cmd);
	bit_clear(TGT_PORT,TGT_XA1);
}

static void pp_data_low(uint8_t data) {
	bit_set(TGT_PORT,TGT_XA0);
	pp_bus_write(data);
	bit_clear(TGT_PORT,TGT_XA0);
}

static void pp_addr_low(uint8_t addr) {
	pp_bus_write(addr);
}

static inline void pp_idle() {
	bit_set(TGT_PORT,TGT_XA1);
	bit_set(TGT_PORT,TGT_XA0);
	pp_bus_write(0x0);
	bit_clear(TGT_PORT,TGT_XA1);
	bit_clear(TGT_PORT,TGT_XA0);
}

static void pp_do_write() {
	_delay_loop_1((uint8_t)(F_CPU/1000000) *  5); // 15 us
	bit_clear(TGT_PORT, TGT_WR);
	_delay_loop_1((uint8_t)(F_CPU/1000000) * 10); // 30 us
	bit_set(TGT_PORT, TGT_WR);
	_delay_loop_1((uint8_t)(F_CPU/1000000) *  5); // 15 us
}

static uint8_t pp_op(uint8_t op, uint8_t data) {
	uint8_t r;
	r = 0xff;
	mdelay(1);
	if (op == OP_IDLE) {
		while (data--) {
			pp_idle();
		}
	} else if (op == OP_SIGNATURE) {
		pp_cmd(0x08);
		pp_addr_low(data);
		r = pp_bus_read();
	} else if (op == OP_CHIP_ERASE) {
		pp_cmd(0x80);
		pp_do_write();
		mdelay(100);
	} else if ( (op == OP_RD_FUSE_L) || (op == OP_RD_FUSE_H)
		   || (op == OP_RD_FUSE_Ext) ) {
		pp_cmd(0x04);
		if (op == OP_RD_FUSE_H) {
			bit_set(TGT_PORT,TGT_BS1);
			bit_set(TGT_PORT,TGT_BS2);
		} else if (op == OP_RD_FUSE_Ext) { 
			bit_set(TGT_PORT,TGT_BS2);
		}
		r = pp_bus_read();
		bit_clear(TGT_PORT,TGT_BS1);
		bit_clear(TGT_PORT,TGT_BS2);
	} else if ( (op == OP_WR_FUSE_L) || (op == OP_WR_FUSE_H)
		   || (op == OP_WR_FUSE_Ext) ) {
		pp_cmd(0x40);
		pp_data_low(data);
		if (op == OP_WR_FUSE_H) {
			bit_set(TGT_PORT,TGT_BS1);
		} else if (op == OP_WR_FUSE_Ext) {
			bit_set(TGT_PORT,TGT_BS2);
		}
		pp_do_write();
		mdelay(100);
		bit_clear(TGT_PORT,TGT_BS2);
		bit_clear(TGT_PORT,TGT_BS1);
	}
	return r;
}
#endif

struct avr_data {
	uint8_t signature[3];
	int8_t  nr_fuse;
	uint8_t fuses[3];
} cur_data;
uint8_t fuses_after[3];

static PROGMEM struct avr_data avr_data[] = {
	/* HSV Type */
	{ { 0x1e, 0x90, 0x07 } , 2 , {  0x6a, 0xff, 0x00}},  // ATtiny13
	{ { 0x1e, 0x91, 0x08 } , 3 , {  0x62, 0xdf, 0xff}},  // ATtiny25
	{ { 0x1e, 0x92, 0x06 } , 3 , {  0x62, 0xdf, 0xff}},  // ATtiny45
	{ { 0x1e, 0x93, 0x0B } , 3 , {  0x62, 0xdf, 0xff}},  // ATtiny85
	{ { 0x1e, 0x91, 0x0B } , 3 , {  0x62, 0xdf, 0xff}},  // ATtiny24
	{ { 0x1e, 0x92, 0x07 } , 3 , {  0x62, 0xdf, 0xff}},  // ATtiny44
	{ { 0x1e, 0x93, 0x0C } , 3 , {  0x62, 0xdf, 0xff}},  // ATtiny84

	/* Tiny2313 Type */
	{ { 0x1e, 0x91, 0x0A } , 3 , {  0x64, 0xdf, 0xff}},  // ATtiny2313
	{ { 0x1e, 0x91, 0x01 } , 1 , {  0xdf, 0x00, 0x00}},  // AT90S2313
	{ { 0x1e, 0x90, 0x01 } , 1 , {  0xdf, 0x00, 0x00}},  // AT90S1200
	//{ { 0x1e, 0x90, 0x01 } , 1 , {  0xde, 0x00, 0x00}},  // AT90S1200A

	/* Tyny26 Type */
	{ { 0x1e, 0x91, 0x09 } , 2 , {  0xe1, 0xf7, 0x00}},  // ATtiny26
	{ { 0x1e, 0x91, 0x0C } , 3 , {  0x62, 0xdf, 0xff}},  // ATtiny261
	{ { 0x1e, 0x92, 0x08 } , 3 , {  0x62, 0xdf, 0xff}},  // ATtiny461
	{ { 0x1e, 0x93, 0x0D } , 3 , {  0x62, 0xdf, 0xff}},  // ATtiny861

	/* Mega8 Type */
	{ { 0x1e, 0x93, 0x07 } , 2 , {  0xe1, 0xd9, 0x00}},  // ATmega8
	{ { 0x1e, 0x92, 0x05 } , 3 , {  0x62, 0xdf, 0xff}},  // ATmega48
	{ { 0x1e, 0x93, 0x0A } , 3 , {  0x62, 0xdf, 0xf9}},  // ATmega88
	{ { 0x1e, 0x94, 0x06 } , 3 , {  0x62, 0xdf, 0xf9}},  // ATmega168
	{ { 0x1e, 0x92, 0x03 } , 1 , {  0xda, 0x00, 0x00}},  // AT90S4433

};

static int8_t find_avr() {
	int8_t i;
	for (i=0; i< sizeof(avr_data)/sizeof(avr_data[0]); i++) {
	    if (  (cur_data.signature[0] == __LPM(avr_data[i].signature+0))
	        &&(cur_data.signature[1] == __LPM(avr_data[i].signature+1))
	        &&(cur_data.signature[2] == __LPM(avr_data[i].signature+2))) {
		return i;
	    }
	}
	return -1;
}

#define PROGRAM_HSV	0
#define PROGRAM_PP	1

static int8_t program(int8_t pt) {
	int i,id,nr_fuse;
	int retry_cnt = 0;
	uint8_t (*op)(uint8_t op, uint8_t data);

#ifndef NO_HSV
	if (pt == PROGRAM_HSV) {
		op = hsv_op;
	} else {
#ifdef NO_PP
		return -1;
#else
		op = pp_op;
#endif
	}
#elif defined(NO_PP)
	return -1;
#else
	op = pp_op;
#endif
retry:
	op(OP_IDLE,10);
	for (i=0; i<3; i++) {
		cur_data.signature[i] = op(OP_SIGNATURE,i);
	}
	id = find_avr();
	if (id < 0) {
		if (cur_data.signature[0] == 0x1e) {
			return -2;
		}
		if (retry_cnt++ < 2) {
			goto retry;
		}
		return -1;
	}
	nr_fuse = __LPM(&avr_data[id].nr_fuse);
	op(OP_CHIP_ERASE,0);
	for (i=0; i<nr_fuse; i++) {
		cur_data.fuses[i] = op(OP_RD_FUSE+i,0);
		op(OP_WR_FUSE+i,__LPM(avr_data[id].fuses+i));
		fuses_after[i] = op(OP_RD_FUSE+i,0);
	}
	for (i=0; i<nr_fuse; i++) {
		if (fuses_after[i] != __LPM(avr_data[id].fuses+i)) {
			return -3 -i;
		}
	}
	return 0;
}

static inline void green_on() {
	bit_set(TGT_XTAL1_PORT, TGT_XTAL1);
}

static inline void green_blink() {
	for (;;) {
		bit_clear(TGT_XTAL1_PORT, TGT_XTAL1);
		ss_sleep(5);
		bit_set(TGT_XTAL1_PORT, TGT_XTAL1);
		ss_sleep(5);
	}
}

static inline void red_blink() {
	for (;;) {
		bit_clear(TGT_OE_PORT, TGT_OE);
		ss_sleep(5);
		bit_set(TGT_OE_PORT, TGT_OE);
		ss_sleep(5);
	}
}

int main(void)
{
    int8_t r;
    /* set system_clock pre scaler 1/8 => 1/1 */
    CLKPR = (1<<CLKPCE);
    CLKPR = 0;

    init_port();
    ss_sleep(30);
    program_enable();
#ifndef NO_HSV
    r = program(PROGRAM_HSV);
    cur_data.nr_fuse = r;
#else
    r = -1;
#endif
    eeprom_write_block (&cur_data, (void *)0, sizeof(cur_data));
    eeprom_write_block (fuses_after, (void *)(sizeof(cur_data))
		    		, sizeof(fuses_after));
    if (r == -1) {
	    r = program(PROGRAM_PP);
    	    cur_data.nr_fuse = r;
    	    eeprom_write_block (&cur_data, (void *)0x10, sizeof(cur_data));
            eeprom_write_block (fuses_after, (void *)(0x10 + sizeof(cur_data))
			    			, sizeof(fuses_after));
    }
    ss_sleep(10);
    program_disable();


    if (r == 0) { // success
	    green_on();
    } else if (r == -2) { // unknown id
	    green_blink();
    } else {
	    red_blink();
    }
    for (;;) {
    }
    return 0;
}
